package ca.yorku.eecs.mack.graffiti;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

/**
 * SpellCheck - class to correct spelling mistakes in words (Android version)
 *
 * @author (c) Scott MacKenzie, 2002-2016
 */
public class SpellCheck
{
	final String MYDEBUG = "MYDEBUG"; // for Log.i messages

	WordFreq[] wf;

	// constructor
	SpellCheck()
	{
	} // all the work done in 'set' and 'get' methods

	public void setDictionary(WordFreq[] wfArg)
	{
		wf = wfArg;
		Arrays.sort(wf, new WordFreqByWord());
	}

	public WordFreq getWordFreq(String wordArg, int freqArg)
	{
		return new WordFreq(wordArg, freqArg);
	}

	// returns true if the given word is in the dictionary
	public boolean wordInDictionary(String word)
	{
		WordFreq dummy = new WordFreq(word, 0); // use freq = 0 (just for the search)
		int n = Arrays.binarySearch(wf, dummy, new WordFreqByWord());
		return n >= 0;
	}

	/**
	 * getCandidates - returns a list of candidate words for the passed word.
	 *
	 * This method is called if the passed word is not in the dictionary.  The candidate list consists of words in the
	 * dictionary that differ from the passed work by MSD = 1, 2, or 3.  The MSD = 1 words are the most likely
	 * candidates since they are closer in "minimum string distance" to the passed word.  The MSD = 2 words are next
	 * most likely, then the MSD = 3 words.
	 *
	 * A final adjustment to the list is to organize the list in two parts.  The first part consists of candidate
	 * words that are equal in length to the passed word.  The second part consists of all other candidate words.
	 * Within each part, the words are placed in frequency order.
	 */
	public String[] getCandidates(String wordArg, int maxArg)
	{
		String[] candidates; // put the candidates here

		ArrayList<WordFreq> list1 = new ArrayList<WordFreq>(); // words with MSD = 1
		ArrayList<WordFreq> list2 = new ArrayList<WordFreq>(); // words with MSD = 2
		ArrayList<WordFreq> list3 = new ArrayList<WordFreq>(); // words with MSD = 3

		// scan the entire dictionary and build a list of candidate words with MSD = 1, 2, or 3
		for (WordFreq wordFreq : wf)
		{
			// can't be a candidate if size difference is greater than three
			if (Math.abs(wordFreq.getWord().length() - wordArg.length()) > 3)
				continue;

			//int distance = (new MSD2(wf[i].getWord(), word)).getMSD();
			int distance = (new MSD(wordFreq.getWord(), wordArg)).getMSD();

			if (distance == 1)
				list1.add(wordFreq);
			else if (distance == 2)
				list2.add(wordFreq);
			else if (distance == 3)
				list3.add(wordFreq);
		}

		WordFreq[] wf1;
		WordFreq[] wf2;
		WordFreq[] wf3;

		wf1 = list1.toArray(new WordFreq[list1.size()]);
		wf2 = list2.toArray(new WordFreq[list2.size()]);
		wf3 = list3.toArray(new WordFreq[list3.size()]);

		Arrays.sort(wf1, new WordFreqByFreq());
		Arrays.sort(wf2, new WordFreqByFreq());
		Arrays.sort(wf3, new WordFreqByFreq());

		// debug
		StringBuilder sb = new StringBuilder("wf1_list: ");
		for (int i = 0; i < wf1.length && i < 10; ++i)
			sb.append(String.format("(%s, %d) ", wf1[i].getWord(), wf1[i].getFreq()));
		//Log.i(MYDEBUG, sb.toString());

		sb = new StringBuilder("wf2_list: ");
		for (int i = 0; i < wf2.length && i < 10; ++i)
			sb.append(String.format("(%s, %d) ", wf2[i].getWord(), wf2[i].getFreq()));
		//Log.i(MYDEBUG, sb.toString());

		sb = new StringBuilder("wf3_list: ");
		for (int i = 0; i < wf3.length && i < 10; ++i)
			sb.append(String.format("(%s, %d) ", wf3[i].getWord(), wf3[i].getFreq()));
		//Log.i(MYDEBUG, sb.toString());

		int total = wf1.length + wf2.length + wf3.length;
		int size = total > maxArg ? maxArg : total;

		if (size == 0)
			return null; // no candidates

		candidates = new String[size];

		for (int j = 0; j < candidates.length; ++j)
		{
			for (int k = 0; k < wf1.length && j < candidates.length; ++k)
				candidates[j++] = wf1[k].getWord();
			for (int k = 0; k < wf2.length && j < candidates.length; ++k)
				candidates[j++] = wf2[k].getWord();
			for (int k = 0; k < wf3.length && j < candidates.length; ++k)
				candidates[j++] = wf3[k].getWord();
		}

		// create a new list with candidates at the front that are equal in length to the entered word
		String[] equalLengthCandidatesFirst = new String[candidates.length];
		int countEqual = 0;

		// put the equal-length candidates at the front (in frequency order)
		for (String s : candidates)
			if (s.length() == wordArg.length())
				equalLengthCandidatesFirst[countEqual++] = s;

		// put the not-equal-length candidates next (in frequency order)
		for (String s : candidates)
			if (s.length() != wordArg.length())
				equalLengthCandidatesFirst[countEqual++] = s;

		return equalLengthCandidatesFirst;
	}

	/**
	 * correctWord - convert the passed word to the correct word
	 *
	 * If the passed word is not in the dictionary AND there is at least one word returned by getCandidates, return
	 * the first word in the candidate list.  Otherwise, return the passed word.
	 */
	public String correctWord(String wordArg)
	{
		String[] temp = getCandidates(wordArg, 10);
		if (!wordInDictionary(wordArg) && temp != null)
			return temp[0];
		else
			return wordArg;
	}

	/**
	 * correctPhrase -- correct a phrase of words
	 *
	 * This method splits a phrase into a String array of words, then compiles an new phrase replacing each word with
	 * the word returned by correctWord.
	 */
	public String correctPhrase(String phraseArg)
	{
		StringBuilder correctedPhrase = new StringBuilder(100);
		for (String w : phraseArg.split("\\s"))
			correctedPhrase.append(String.format("%s ", correctWord(w)));
		return correctedPhrase.toString().trim();
	}

	// WordFreq - class to hold a word (String) and its frequency (int) from a corpus
	public class WordFreq
	{
		String word;
		int freq;

		WordFreq(String wordArg, int freqArg)
		{
			word = wordArg;
			freq = freqArg;
		}

		public String getWord()
		{
			return word;
		}

		public int getFreq()
		{
			return freq;
		}

		public String toString()
		{
			return word + " : " + freq;
		}
	}

	// WordFreqByWord -- a class that implements Comparator for sorting WordFreq objects by words
	public class WordFreqByWord implements Comparator<WordFreq>
	{
		public int compare(WordFreq wf1, WordFreq wf2)
		{
			String s1 = wf1.getWord();
			String s2 = wf2.getWord();
			return s1.compareTo(s2);
		}
	}

	// WordFreqByFreq - a class that implements Comparator for sorting WordFreq objects by frequency
	public class WordFreqByFreq implements Comparator<WordFreq>
	{
		public int compare(WordFreq wf1, WordFreq wf2)
		{
			int i1 = wf1.getFreq();
			int i2 = wf2.getFreq();
			return i2 - i1; // descending
		}
	}
}
